/*
    bysel (get elements by selector)
    version 1.0 Copyright (C) 2007 Tagata Akira

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
    02110-1301 USA

*/

// getElementsBySelector

// yacc
//
//	combinator
//	  : PLUS S*
//	  | GREATER S*
//	  | S
//	  ;
//
//
//	selector
//	  : simple_selector [ combinator simple_selector ]*
//	  ;
//	simple_selector
//	  : element_name [ HASH | class | attrib | pseudo ]*
//	  | [ HASH | class | attrib | pseudo ]+
//	  ;
//	class
//	  : '.' IDENT
//	  ;
//	element_name
//	  : IDENT | '*'
//	  ;
//	attrib
//	  : '[' S* IDENT S* [ [ '=' | INCLUDES | DASHMATCH ] S*
//	    [ IDENT | STRING ] S* ]? ']'
//	  ;
//	pseudo
//	  : ':' [ IDENT | FUNCTION S* IDENT? S* ')' ]
//	  ;
//
// flex
// %option case-insensitive
//
//	{s}			{return S;}
//	"~="			{return INCLUDES;}
//	"|="			{return DASHMATCH;}
//	{w}"+"			{return PLUS;}
//	{w}">"			{return GREATER;}
//	{string}		{return STRING;}
//	{ident}			{return IDENT;}
//	{ident}"("		{return FUNCTION;}
//
//	h		[0-9a-f]
//	s		[ \t\r\n\f]+
//	w		{s}?
//	unicode		\\{h}{1,6}(\r\n|[ \t\r\n\f])?
//	escape		{unicode}|\\[^\r\n\f0-9a-f]
//	nonascii	[\200-\377]
//	nmstart		[_a-z]|{nonascii}|{escape}
//	nmchar		[_a-z0-9-]|{nonascii}|{escape}

//	ident		-?{nmstart}{nmchar}*
//	name		{nmchar}+

//	nl		\n|\r\n|\r|\f
//	string1		\"([^\n\r\f\\"]|\\{nl}|{escape})*\"
//	string2		\'([^\n\r\f\\']|\\{nl}|{escape})*\'
//	string		{string1}|{string2}

_fss.getElementsBySelector = function (selector,top) {

	if (typeof selector != 'string' && !selector.length) return;

	////////////////////////////////////////////////////////////////
	function reMatch(constRe,buff) {
		var ret = null;
		if (!(constRe in _fss.constBysel.reCache))
			_fss.constBysel.reCache[constRe] = new RegExp(constRe,'ig');
		var re = _fss.constBysel.reCache[constRe];
		re.lastIndex = buff.lastIndex;
		var matches = re.exec(buff.str);
		if (matches && matches.index == buff.lastIndex) {
			ret = matches;
			buff.lastIndex = re.lastIndex;
		}
		return ret;
	}
	////////////////////////////////////////////////////////////////
	function Hcap(buff) {
//alert('buff.lastIndex:'+buff.lastIndex);

		var ret = null;

		// matched array
		var mHASH, mClassName, mAttrib,mPseudo;

		if ((mHASH = reMatch(_fss.constBysel.HASH,buff)) != null) {
			ret = { HASH: mHASH[0].substring(1) };
		}
		else if ((mClassName = reMatch(_fss.constBysel.className,buff)) != null) {
			ret = { className: mClassName[0].substring(1) };
		}
		else if ((mAttrib = reMatch(_fss.constBysel.attrib,buff)) != null) {
			var indexAttrib = 2, indexEqual = 11, indexValue = 13;
			if (mAttrib[indexAttrib]) {
				ret = { attrib: mAttrib[indexAttrib] };
				if (mAttrib[indexEqual] && mAttrib[indexValue]) {
					if (mAttrib[indexEqual] == '=') {
						ret.EXACTLY = mAttrib[indexValue];
					}
					else if (mAttrib[indexEqual] == '~=') {
						ret.INCLUDES = mAttrib[indexValue];
					}
					else if (mAttrib[indexEqual] == '|=') {
						ret.DASHMATCH = mAttrib[indexValue];
					}
				}
			}
		}
		else if ((mPseudo = reMatch(_fss.constBysel.pseudo,buff)) != null) {
			var indexPseudo = 2, indexFuncArg = 1, indexFunc = 2, indexParen = 9, indexArg = 11;
//alert('buff.str:'+buff.str+', buff.lastIndex:'+buff.lastIndex+', mPseudo.length:'+mPseudo.length+', _fss.constBysel.pseudo:'+_fss.constBysel.pseudo);
//alert('mPseudo[indexPseudo]:'+mPseudo[indexPseudo]+', mPseudo[indexFuncArg]:'+mPseudo[indexFuncArg]+', mPseudo[indexFunc]:'+mPseudo[indexFunc]+', mPseudo[indexParen]:'+mPseudo[indexParen]+', mPseudo[indexArg]:'+mPseudo[indexArg]);
//			if (mPseudo[indexFuncArg] && mPseudo[indexFunc] && mPseudo[indexArg]) {
//				ret = { pseudo: mPseudo[indexFuncArg], func: mPseudo[indexFunc], arg: mPseudo[indexArg]};
			if (mPseudo[indexParen]) {
				ret = { pseudo: mPseudo[indexFuncArg], func: mPseudo[indexPseudo], arg: mPseudo[indexArg]};
			}
			else if (mPseudo[indexPseudo]) {
				ret = { pseudo: mPseudo[indexPseudo] };
			}
		}
//alert('ret.HASH:'+ret.HASH+', ret.className:'+ret.className);
		return ret;
	}
	////////////////////////////////////////////////////////////////
	function simpleSelector(buff) {
		var reElementName = RegExp(_fss.constBysel.elementName,'ig');
		reElementName.lastIndex = buff.lastIndex;
		var matches = reElementName.exec(buff.str);

		var elementName;
		if (matches && matches.index == buff.lastIndex) {
			elementName = matches[0];
			buff.lastIndex = reElementName.lastIndex;
		}

		var ret = { elementName: elementName };
		var x = ret;
		while (x.hcap = Hcap(buff))
			x = x.hcap;
////////////////////////////////////////////////////////////////
//var xx = 'ret.elementName:'+ret.elementName+',';
//for (var yy = ret.hcap; yy; yy = yy.hcap) {
//	xx += '\n'+'yy.HASH:'+yy.HASH+', yy.className:'+yy.className+', yy.attrib:'+yy.attrib+', yy.pseudo:'+yy.pseudo;
//}
//alert(xx);
////////////////////////////////////////////////////////////////
		return ret;
	}
	////////////////////////////////////////////////////////////////
	function findCombinator(buff) {
		var reCombinator = RegExp(_fss.constBysel.combinator,'g');
		reCombinator.lastIndex = buff.lastIndex;
		var combinator;
		var matches = reCombinator.exec(buff.str);
		if (matches && matches.index == buff.lastIndex) {
			if (matches[0].indexOf(_fss.constBysel.PLUS) != -1)
				combinator = _fss.constBysel.PLUS;
			else if (matches[0].indexOf(_fss.constBysel.GREATER) != -1)
				combinator = _fss.constBysel.GREATER;
			else
				combinator = ' ';
		}
		buff.lastIndex = reCombinator.lastIndex;
		return combinator;
	}
	////////////////////////////////////////////////////////////////
	function attribValue(value) {
		if (value.charAt(0) == '\'' || value.charAt(0) == '\"') {
			return value.substring(1,value.length-1);
		}
		return value;
	}
	////////////////////////////////////////////////////////////////
	function eltMatch(elt,ss) {
		var eltMatched = true;
		if (ss.elementName && ss.elementName != '*')
			eltMatched = elt.nodeName.toLowerCase() == ss.elementName;
		for (var hcap = ss.hcap; eltMatched && hcap; hcap = hcap.hacp) {
			if (hcap.HASH) { // id
				if (elt.id == hcap.HASH)
					continue;
				else
					eltMatched = false;
			}
			else if (hcap.className) { // class
				if (elt.className) {
//					var reClassName = new RegExp('\\b'+hcap.className+'\\b');
					var reClassName = new RegExp('(^|[^_A-Za-z0-9-])'+hcap.className+'([^_A-Za-z0-9-]|$)');
					if (reClassName.test(elt.className))
						continue;
					else
						eltMatched = false;
				}
				else
					eltMatched = false;
			}
			else if (hcap.attrib) { // attr
//alert('hcap.attrib:'+hcap.attrib+', hcap.EXACTLY:'+hcap.EXACTLY+', hcap.INCLUDES:'+hcap.INCLUDES+', hcap.DASHMATCH:'+hcap.DASHMATCH);
				if (hcap.EXACTLY != undefined) { // =
					eltMatched = elt[hcap.attrib] == attribValue(hcap.EXACTLY);
				} else if (hcap.INCLUDES != undefined) { // ~=
					var reINCLUDES = new RegExp('\\b'+attribValue(hcap.INCLUDES)+'\\b');
					eltMatched = reINCLUDES.test(elt[hcap.attrib]);
				} else if (hcap.DASHMATCH != undefined) { // |=
					var reDASHMATCH = new RegExp('^'+attribValue(hcap.DASHMATCH)+'(-.*)?$');
					eltMatched = reDASHMATCH.test(elt[hcap.attrib]);

				} else { // only attrib
					eltMatched = hcap.attrib in elt;
				}
			}
			else if (hcap.pseudo) {
//alert('hcap.pseudo:'+hcap.pseudo+', hcap.func:'+hcap.func+', hcap.arg'+hcap.arg);
				if (hcap.func != undefined && hcap.arg != undefined) {
					if (hcap.func == 'lang') {
//						hcap.pseudo; // FUNC(ARG)
//						hcap.func; // lang
						var reLang = new RegExp('^'+hcap.arg+'(-.*)?$');
						// search lang upward
						var x = elt;
						do {
							if (x.lang) {
								eltMatched = relLang.test(x.lang);
								break;
							}
						} while (x = x.parentNode);
						if (!x)
							eltMatched = false;
					}
					else
						eltMatched = false;
				}
				else { // no func
					if (hcap.pseudo == 'first-child')
						eltMatched = !elt.previousSibling;
					else
						eltMatched = false;
				}
			}
		}
		return eltMatched;
	}
	////////////////////////////////////////////////////////////////
	function getElements(lastElements,combinator,ss) {

		var elements = new Array;

//alert('combinator:'+combinator+', lastElements.length:'+lastElements.length+', ss.elementName:'+ss.elementName+', ss.hcap.HASH:'+ss.hcap.HASH+', ss.hcap.className:'+ss.hcap.className);
		for (var i = 0; i < lastElements.length; i++) {
			if (combinator == ' ') { // descendant
				var tagName = '*';
				if (ss.elementName && ss.elementName != '*')
					tagName = ss.elementName;
				var elts;
				if (ss.hcap && ss.hcap.HASH) {
					elts = new Array;
					var eltIdentified = document.getElementById(ss.hcap.HASH);
					if (eltIdentified)
						elts.push(eltIdentified);
				}
				else
					elts = lastElements[i].getElementsByTagName(tagName);
//alert('elts:'+elts+', elts.length:'+elts.length);
				if (elts && elts.length) {
					for (j = 0; j < elts.length; j++) {
						if (eltMatch(elts[j],ss))
							elements.push(elts[j]);
					}
				}
			}
			else if (combinator == _fss.constBysel.PLUS) { // next sibling
				var elt = lastElements[i].nextSibling;
				if (eltMatch(elt,ss))
					elements.push(elt);
			}
			else if (combinator == _fss.constBysel.GREATER) { // child
				for (x = lastElements[i].firstChild; x; x = x.nextSibling) {
					if (eltMatch(x,ss))
						elements.push(x);
				}
			}
		}
		return elements;
	}
	////////////////////////////////////////////////////////////////


	var buff = { str: selector, lastIndex: 0 };
	var combinator = ' ';
	var lastElements = [top || document];
	var ss;
	var ret;
	do {
		if (!(ss = simpleSelector(buff))) break;
		lastElements = getElements(lastElements,combinator,ss);
		ret = lastElements;
		combinator = findCombinator(buff);
//alert('lastElements.length:'+lastElements.length+', combinator:|'+combinator+'|');
	} while (lastElements.length && combinator);

	_fss.getElementsBySelector.prototype._fssInsertUrl();
	return ret;
};

_fss.getElementsBySelector.prototype = new _fss('bysel','get elements by selector','1.0alpha','2007 Tagata Akira','http://www.tagata.org/software/');

//-fxref-// _fss.getElementsBySelector: _fss

_fss.getElementsBySelectors = function (selectors, top) {

	var elts = new Array;
	if (selectors instanceof Array) {
		for (var i = 0; i < selectors.length; i++)
			elts[i] = _fss.getElementsBySelector(selectors[i],top);
	}
	else if (typeof selectors == 'string') {
		elts[0] = _fss.getElementsBySelector(selectors,top);
	}
	return elts;
}
//-fxref-// _fss.getElementsBySelectors: _fss.getElementsBySelector

_fss.constBysel = { _init: function () {
	this.unicode = '\\[0-9a-f]{1,6}(\r\n|[ \t\r\n\f])?';
	this.nonascii = '[\200-\377]';
	this.escape = this.unicode + '|' + '\\[^\r\n\f0-9a-f]';
	this.nmstart = '[_a-z]' + '|' + this.nonascii + '|' + '('+ this.escape + ')';
	this.nmchar = '[_a-z0-9-]' + '|' + this.nonascii + '|' + '('+ this.escape + ')';
	this.ident = '-?' + '(' + this.nmstart + ')' + '(' + this.nmchar + ')' + '*';
	this.elementName = this.ident + '|' + '\\*';
	this.HASH = '#' + '(' + this.nmchar + ')+';
	this.className = '\\.' + this.ident;

	this.S = '(' + '[ \t\r\n\f]+' + ')';
	this.W = '(' + this.S + '?' + ')';

	this.PLUS = '+';
	this.GREATER = '>';

	this.nl = '(\n|\r\n|\r|\f)';
	this.string1 = '\\\"' + '(' + '[^\n\r\f\\\\\"]' + '|' + '\\\\' + this.nl + '|' + '(' + this.escape + ')' + ')' + '*' + '\\\"';
	this.string2 = '\\\'' + '(' + '[^\n\r\f\\\\\']' + '|' + '\\\\' + this.nl + '|' + '(' + this.escape + ')' + ')' + '*' + '\\\'';
	this.string = '(' + this.string1 + '|' + this.string2 + ')';

	this.attrib = '\\[' + this.S + '*' + '(' + this.ident + ')' + this.S + '*' + '(' + '(=|~=|\\|=)' + this.S + '*' + '(' + this.ident + '|' + this.string + ')' + this.S + '*' + ')' + '?' + '\\]';
//alert('this.attrib:'+this.attrib);
//	this.pseudo = ':' + '(' + '(' + this.ident + ')' + '|' + '(' + this.ident + ')' + '\\(' + this.S + '*' + '(' + this.ident + ')' + '?' + this.S + '*' +'\\)' + ')';
	this.pseudo = ':' + '('
		+ '(' + this.ident + ')' + '('
		+ '\\(' + this.S + '*' + '(' + this.ident + ')' + '?' + this.S + '*' +'\\)' + ')?' + ')';
//alert('this.pseudo:'+this.pseudo);
	this.combinator =
		this.W + '\\' + this.PLUS + this.S + '*' + '|' +
		this.W + this.GREATER + this.S + '*' + '|' +
		this.S;
	this.reCache = new Object;
	}
};
_fss.constBysel._init();